#!/usr/bin/perl

# This a perl script that reads a xyz file (TINKER format)
# and four integers that correspond to four atoms
# and returns the dihedral value
#
# The subroutine that makes the computation is a modification
# of the original tinker subroutine geometry.f
# TINKER can be downloaded from http://dasher.wustl.edu/tinker/
#
# Script written by Athanassios Stavrakoudis ( astavrak@cc.uoi.gr)
# or ( http://www.econ.uoi.gr/stavrakoudis/) and can be freely distributed
# and/or modifed.
#
# To run the script type dihedral.pl foo.xyz 11 7 13 12
# or whatever fits to your system
#


use Math::BigFloat;
use Math::Trig;

# read a xyz file
$file = $ARGV[0];
$at1  = $ARGV[1];
$at2  = $ARGV[2];
$at3  = $ARGV[3];
$at4  = $ARGV[4];

open (XYZ,"<$file") || die "  $file could not be found ";
@xyz = (<XYZ>);
close (XYZ);

# initialize
my @n    = ();
my @type = ();
my @x    = ();
my @y    = ();
my @z    = ();

# take xyz coordinates
for ($i=1; $i<=$#xyz; ++$i) {
    ($null, $n[$i], $type[$i], $x[$i], $y[$i], $z[$i]) = split (/\s+/, $xyz[$i]);
    if ($n[$i] == $at1) {
        $x1 = $x[$i];
        $y1 = $y[$i];
        $z1 = $z[$i];
    }
    if ($n[$i] == $at2) {
        $x2 = $x[$i];
        $y2 = $y[$i];
        $z2 = $z[$i];
    }
    if ($n[$i] == $at3) {
        $x3 = $x[$i];
        $y3 = $y[$i];
        $z3 = $z[$i];
    }
    if ($n[$i] == $at4) {
        $x4 = $x[$i];
        $y4 = $y[$i];
        $z4 = $z[$i];
    }
}


# print result

$dih = dihedral($x1,$y1,$z1,$x2,$y2,$z2,$x3,$y3,$z3,$x4,$y4,$z4);
printf ("%8.1f\n",$dih);



# subroutine (taken from tinker subroutine geometry.f) to compute a dihedral
sub dihedral {

    use Math::BigFloat;
    use Math::Trig;

    my $radian = 57.29577951308232088;
    ($x1,$y1,$z1, $x2,$y2,$z2, $x3,$y3,$z3, $x4,$y4,$z4) = @_;

    my $xba = $x2 - $x1;
    my $yba = $y2 - $y1;
    my $zba = $z2 - $z1;
    my $xcb = $x3 - $x2;
    my $ycb = $y3 - $y2;
    my $zcb = $z3 - $z2;
    my $xdc = $x4 - $x3;
    my $ydc = $y4 - $y3;
    my $zdc = $z4 - $z3;
    my $xt = $yba*$zcb - $ycb*$zba;
    my $yt = $xcb*$zba - $xba*$zcb;
    my $zt = $xba*$ycb - $xcb*$yba;
    my $xu = $ycb*$zdc - $ydc*$zcb;
    my $yu = $xdc*$zcb - $xcb*$zdc;
    my $zu = $xcb*$ydc - $xdc*$ycb;
    my $rt2 = $xt*$xt + $yt*$yt + $zt*$zt;
    my $ru2 = $xu*$xu + $yu*$yu + $zu*$zu;
    my $rtru = sqrt($rt2 * $ru2);

    if ($rtru != 0) {
        my  $cosine = ($xt*$xu + $yt*$yu + $zt*$zu) / $rtru;
        my  $cosine = min(1,max(-1,$cosine));
        my  $dihedral = $radian * acos($cosine);
        my  $sign = $xba*$xu + $yba*$yu + $zba*$zu;
        if ( $sign < 0 ) { $dihedral = -$dihedral; }
       return $dihedral;
    }

    return $dihedral;
}

#
# some auxilary subroutines
#
sub acos {
    $acos = atan2( sqrt(1 - $_[0] * $_[0]), $_[0] );
    return $acos;
}

sub max {
    my $max = shift(@_);
    foreach $_ (@_) {
        if ($_ > $max) { $max = $_; }
    }
return $max;
}

sub min {
    my $min = shift(@_);
    foreach $_ (@_) {
        if ($_ < $min ) { $min = $_; }
    }
    return $min;
}
